// Project: eBus Connector
//  Module: Console Wrapper
//  Author: Thomas Ternyi (thomas@terenyi.at)

using System;
using System.Reflection;

[assembly: AssemblyTitle("eBus Connector Console Wrapper")]
[assembly: AssemblyCompany("Thomas Ternyi")]
[assembly: AssemblyProduct("eBus Connector")]
[assembly: AssemblyCopyright("Copyright  2012")]
[assembly: AssemblyVersion("1.3.0.0")]
[assembly: AssemblyFileVersion("1.3.0.0")]

namespace eBusConnector
{
    /// <summary>
    /// This class contains the console wrapper.
    /// </summary>
    class ConsoleWrapper
    {
        /// <summary>
        /// The eBus Connector.
        /// </summary>
        private static eBusConnector eBus;
        
        /// <summary>
        /// This is the main entry point for the application.
        /// </summary>
        /// <param name="args">Command line parameters.</param>
        static void Main(string[] args)
        {
            // Print the program title
            Title();

            // Declare variables
            string serialPort;
            string localUDPPort;
            string host;
            string remoteUDPPort;
            bool listenWhileSending;
            bool stopwatch;

            try
            {
                // Read the settings
                if(args.Length != 4 && args.Length != 5 && args.Length != 6)
                    throw new Exception();

                serialPort = args[0];
                localUDPPort = args[1];
                host = args[2];
                remoteUDPPort = args[3];

                listenWhileSending = false;
                stopwatch = false;

                if(args.Length == 5 || args.Length == 6)
                {
                    if(args[4] == "-lws")
                        listenWhileSending = true;

                    if(args[4] == "-sw")
                        stopwatch = true;
                }

                if(args.Length == 6)
                {
                    if(args[5] == "-lws")
                        listenWhileSending = true;

                    if(args[5] == "-sw")
                        stopwatch = true;
                }
                
                // Check the settings
                if(int.Parse(localUDPPort) <= 0 || int.Parse(localUDPPort) > 65535)
                    throw new Exception();

                if(int.Parse(remoteUDPPort) <= 0 || int.Parse(remoteUDPPort) > 65535)
                    throw new Exception();
            }
            catch(Exception)
            {
                // Show the usage information if the parameters are not correct
                Usage();
                return;
            }

            try
            {
                // Create and configure the eBus Connector
                eBus = new eBusConnector(serialPort, int.Parse(localUDPPort), int.Parse(remoteUDPPort), host, listenWhileSending, stopwatch);
                eBus.Log += Log;

                // Start the eBus Connector
                eBus.Start();

                // Wait for the user to press RETURN
                Console.WriteLine("\n*** PRESS <RETURN> TO EXIT ***\n\n");
                Console.ReadLine();

                // Stop the eBus Connector
                eBus.Stop();
            }
            catch(Exception e)
            {
                // Log errors
                Console.WriteLine("ERROR: " + e.Message);
                Console.ReadLine();
            }
        }

        /// <summary>
        /// Logs a message.
        /// </summary>
        /// <param name="message">The message to log.</param>
        private static void Log(string message)
        {
            Console.WriteLine(message);
        }

        /// <summary>
        /// Prints the program title.
        /// </summary>
        private static void Title()
        {
            Console.Clear();

            Console.WriteLine("+---------------------+");
            Console.WriteLine("| eBus Connector v1.3 |");
            Console.WriteLine("+---------------------+");
            Console.WriteLine("| Thomas Ternyi      |");
            Console.WriteLine("| (thomas@terenyi.at) |");
            Console.WriteLine("+---------------------+");
            Console.WriteLine("");
        }

        /// <summary>
        /// Prints the usage information.
        /// </summary>
        private static void Usage()
        {
            Console.WriteLine("USAGE:");
            Console.WriteLine("");
            Console.WriteLine("eBusConnectorConsole SerialPort LocalUDPPort Host RemoteUDPPort -lws -sw");
            Console.WriteLine("");
            Console.WriteLine("SerialPort:    The name of the serial port used to communicate with the eBus.");
            Console.WriteLine("LocalUDPPort:  The local UDP port to listen on for commands.");
            Console.WriteLine("Host:          The remote host to send the results to.");
            Console.WriteLine("RemoteUDPPort: The remote UDP port to send the results to.");
            Console.WriteLine("RemoteUDPPort: The remote UDP port to send the results to.");
            Console.WriteLine("-lws:          Enables the \"listen while sending\" function (optional).");
            Console.WriteLine("-sw:           Enables the stopwatch while sending (optional).");
            Console.WriteLine("");
            Console.WriteLine("EXAMPLE: eBusConnectorConsole COM10 8812 192.168.0.1 8813 -lws");
            Console.WriteLine("");
        }
    }
}
