// Project: eBus Connector
//  Module: Library
//  Author: Thomas Ternyi (thomas@terenyi.at)

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO.Ports;
using System.Net;
using System.Net.Sockets;
using System.Reflection;
using System.Text;
using System.Threading;

[assembly: AssemblyTitle("eBus Connector Library")]
[assembly: AssemblyCompany("Thomas Ternyi")]
[assembly: AssemblyProduct("eBus Connector")]
[assembly: AssemblyCopyright("Copyright  2012")]
[assembly: AssemblyVersion("1.3.0.0")]
[assembly: AssemblyFileVersion("1.3.0.0")]

namespace eBusConnector
{
    /// <summary>
    /// This class contains the eBus Connector.
    /// </summary>
    public class eBusConnector
    {
        /// <summary>
        /// Logs a message.
        /// </summary>
        /// <param name="message">The message to log.</param>
        public delegate void LogHandler(string message);

        /// <summary>
        /// The logging event.
        /// </summary>
        public event LogHandler Log;

        /// <summary>
        /// The name of the serial port used to communicate with the eBus.
        /// </summary>
        private string SerialPortName;

        /// <summary>
        /// The local UDP port to listen on for commands.
        /// </summary>
        private int LocalUDPPort;

        /// <summary>
        /// The remote UDP port to send the results to.
        /// </summary>
        private int RemoteUDPPort;

        /// <summary>
        /// The remote host to send the results to.
        /// </summary>
        private string Host;

        /// <summary>
        /// The serial port used to communicate with the eBus.
        /// </summary>
        private SerialPort SerialPort;

        /// <summary>
        /// Indicates whether the "listen while sending" function is enabled or disabled.
        /// </summary>
        private bool ListenWhileSending;

        /// <summary>
        /// Indicates whether the stopwatch is enabled or disabled while sending.
        /// </summary>
        private bool Stopwatch;

        /// <summary>
        /// The buffer for received eBus data.
        /// </summary>
        private volatile string Buffer;

        /// <summary>
        /// The previous eBus message that has been received.
        /// </summary>
        private volatile string LastMessage;

        /// <summary>
        /// The eBus messages queued for sending.
        /// </summary>
        private volatile List<string> SendQueue;

        /// <summary>
        /// The thread for receiving UDP messages.
        /// </summary>
        private Thread ReceiverThread;

        /// <summary>
        /// The socket to receive UDP messages.
        /// </summary>
        private UdpClient Receiver;

        /// <summary>
        /// The socket to send UDP messages.
        /// </summary>
        private UdpClient Sender;

        /// <summary>
        /// The endpoint for receiving UDP messages.
        /// </summary>
        private IPEndPoint ReceiverEndPoint;

        /// <summary>
        /// The endpoint for sending UDP messages.
        /// </summary>
        private IPEndPoint SenderEndPoint;

        /// <summary>
        /// Constructs a new eBus Connector.
        /// </summary>
        /// <param name="comPort">The name of the serial port used to communicate with the eBus.</param>
        /// <param name="localUDPPort">The local UDP port to listen on for commands.</param>
        /// <param name="remoteUDPPort">The remote UDP port to send the results to.</param>
        /// <param name="host">The remote host to send the results to.</param>
        /// <param name="listenWhileSending">Indicates whether the "listen while sending" function is enabled or disabled.</param>
        /// <param name="stopwatch">Indicates whether the stopwatch is enabled or disabled while sending.</param>
        public eBusConnector(string serialPortName, int localUDPPort, int remoteUDPPort, string host, bool listenWhileSending, bool stopwatch)
        {
            // Save the parameters
            SerialPortName = serialPortName;
            LocalUDPPort = localUDPPort;
            RemoteUDPPort = remoteUDPPort;
            Host = host;
            ListenWhileSending = listenWhileSending;
            Stopwatch = stopwatch;
        }

        /// <summary>
        /// Starts the eBus Connector.
        /// </summary>
        public void Start()
        {
            // Start the eBus Connector
            Log("Starting eBus Connector...");

            // Resolve the destination host
            string ip = "";

            foreach(IPAddress currentIP in Dns.GetHostEntry(Host).AddressList)
                if(!currentIP.ToString().Contains(":"))
                    ip = currentIP.ToString();

            if(ip == "")
                throw new Exception("Could not resolve " + Host + ".");

            // Start the UDP receiver
            Log("Starting UDP receiver on port " + LocalUDPPort + "...");
            ReceiverEndPoint = new IPEndPoint(IPAddress.Any, LocalUDPPort);
            Receiver = new UdpClient(ReceiverEndPoint);
            ReceiverThread = new Thread(this.ReceiveUDP);
            ReceiverThread.Start();
            Log("UDP receiver started.");

            // Start the UDP sender
            Log("Starting UDP sender to " + Host + ":" + RemoteUDPPort + "...");
            SenderEndPoint = new IPEndPoint(IPAddress.Parse(ip), RemoteUDPPort);
            Sender = new UdpClient();
            Log("UDP sender started.");

            // Clear the buffer
            Buffer = "";
            LastMessage = "";

            // Initialize the send queue
            SendQueue = new List<string>();

            // Create the serial port
            SerialPort = new SerialPort(SerialPortName, 2400, Parity.None, 8, StopBits.One);

            // Add the event handler for receiving data
            SerialPort.DataReceived += SerialDataReceived;

            // Open the serial port
            Log("Opening " + SerialPortName + "...");
            SerialPort.Open();
            SerialPort.ReceivedBytesThreshold = 3;
            Log(SerialPortName + " opened.");

            // Show the status of the "listen while sending" function
            if(ListenWhileSending)
                Log("Listen while sending enabled.");
            else
                Log("Listen while sending disabled.");

            // Show the status of the stopwatch
            if(Stopwatch)
                Log("Stopwatch enabled.");
            else
                Log("Stopwatch disabled.");

            // Finished starting the eBus Connector
            Log("eBus Connector started.\n");
        }

        /// <summary>
        /// Stops the eBus Connector.
        /// </summary>
        public void Stop()
        {
            // Stop the eBus Connector
            Log("Stopping eBus Connector...");

            // Close the serial port
            Log("Closing " + SerialPortName + "...");
            SerialPort.Close();
            Log(SerialPortName + " closed.");

            // Stop the UDP receiver
            Log("Stopping UDP receiver...");
            Receiver.Close();
            ReceiverThread.Abort();
            Log("UDP receiver stopped.");

            // Stop the UDP sender
            Log("Stopping UDP sender...");
            Sender.Close();
            Log("UDP sender stopped.");

            // Finished stopping the eBus Connector
            Log("eBus Connector stopped.\n");
        }

        /// <summary>
        /// Receives data from the serial port.
        /// </summary>
        /// <param name="sender">The sender of the event.</param>
        /// <param name="e">The event data.</param>
        private void SerialDataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            try
            {
                // Read the new data from the serial port
                int bytes = SerialPort.BytesToRead;
                byte[] data = new byte[bytes];
                SerialPort.Read(data, 0, bytes);

                // Append the new data to the buffer
                lock(Buffer)
                    Buffer = Buffer + ByteToHex(data);

                // Process any complete eBus messages that have been received
                while(Buffer.Contains("AA "))
                {
                    // Extract the message from the buffer
                    string message = Buffer.Substring(0, Buffer.IndexOf("AA "));

                    // Remove the message from the buffer
                    lock(Buffer)
                        Buffer = Buffer.Substring(Buffer.IndexOf("AA ") + 3);

                    // Process the message
                    ProcessMessage(message);
                }
            }
            catch(Exception ex)
            {
                // Log errors
                Log("ERROR: " + ex.Message);
            }
        }

        /// <summary>
        /// Processes a complete eBus message.
        /// </summary>
        /// <param name="message">The eBus message to process.</param>
        private void ProcessMessage(string message)
        {
            try
            {
                // Log the message if it has at least the minimum length to be valid
                if(message.Length >= 17)
                    Log("[RECEIVED] " + message);

                // Send the next queued message if no bytes are waiting to be read from the serial port,
                // the buffer contains no more data and if the last message has been an auto SYN
                if(Buffer == "" && LastMessage == "" && SerialPort.BytesToRead == 0)
                {
                    // Lock the send queue
                    lock(SendQueue)
                    {
                        // Process the next message on the queue
                        if(SendQueue.Count != 0)
                        {
                            // Send the message
                            bool sendSucessful = SendMessage(SendQueue[0]);

                            // Remove the message from the queue if it was succesfully sent
                            if(sendSucessful)
                                SendQueue.RemoveAt(0);
                        }
                    }
                }

                // Forward the message if it has at least the minimum length to be valid
                if(message.Length >= 17)
                    SendUDP(message);
               
                // Save the message
                lock(LastMessage)
                    LastMessage = message;
            }
            catch(Exception e)
            {
                // Log errors
                Log("ERROR: " + e.Message);
            }
        }

        /// <summary>
        /// Sends a message to the eBus.
        /// </summary>
        /// <param name="message">The eBus message to send.</param>
        /// <returns>True if the message was successfully sent, false otherwise.</returns>
        private bool SendMessage(string message)
        {
            try
            {
                // Declare the stopwatch
                Stopwatch sw = null;

                // Convert the message
                byte[] buffer = HexToByte(message);

                // Send the whole message at once if "listen while sending" is disabled
                if(!ListenWhileSending)
                {
                    // Start the stopwatch
                    sw = System.Diagnostics.Stopwatch.StartNew();

                    // Write the whole message to the serial port
                    SerialPort.Write(buffer, 0, buffer.Length);

                    // Stop the stopwatch
                    sw.Stop();
                }
                // Send the whole message byte by byte if "listen while sending" is enabled
                else
                {
                    // Start the stopwatch
                    sw = System.Diagnostics.Stopwatch.StartNew();

                    // Write each byte of the message to the serial port
                    for(int i = 0; i < buffer.Length; i++)
                    {
                        // Send the next byte
                        SerialPort.Write(buffer, i, 1);

                        // Wait for the byte to be read back
                        while(SerialPort.BytesToRead < 1)
                            Thread.SpinWait(1);

                        // Read the byte from the serial port
                        int bytes = SerialPort.BytesToRead;
                        byte[] data = new byte[bytes];
                        SerialPort.Read(data, 0, bytes);

                        // Abort if the sent byte does not match the received data or if more than one byte was received (unless 00 was received after sending the last byte)
                        if(data[0] != buffer[i] || (data.Length > 1 && i < buffer.Length - 1) || (data.Length > 1 && i == buffer.Length - 1 && ByteToHex(data[1]) != "00"))
                        {
                            // Stop the stopwatch
                            sw.Stop();

                            // Log the send error
                            Log("[SENDING ] " + ByteToHex(buffer));
                            Log("[SEND ERR] Expected " + ByteToHex(buffer[i]) + ", received " + ByteToHex(data));

                            // Log the elapsed time
                            if(Stopwatch)
                                Log("[TIMER   ] " + sw.ElapsedMilliseconds + "ms");

                            // Add the received byte(s) to the buffer if the error occurred at the first byte and does not contain a SYN symbol
                            if(i == 0 && !ByteToHex(data).Contains("AA"))
                            {
                                lock(Buffer)
                                    Buffer = Buffer + ByteToHex(data);
                            }
                            // Otherwise clear the buffer
                            else
                            {
                                lock(Buffer)
                                    Buffer = "";

                                lock(LastMessage)
                                    LastMessage = "ERROR";
                            }

                            // Message was not sent successfully
                            return false;
                        }

                        // Append the new data to the buffer
                        lock(Buffer)
                            Buffer = Buffer + ByteToHex(data);
                    }

                    // Stop the stopwatch
                    sw.Stop();
                }

                // Log the sent message
                Log("[SENT    ] " + ByteToHex(buffer));

                // Log the elapsed time
                if(Stopwatch)
                    Log("[TIMER   ] " + sw.ElapsedMilliseconds + "ms");
                
                // Message was successfully sent
                return true;
            }
            catch(Exception e)
            {
                // Log errors
                Log("ERROR: " + e.Message);

                // Message was not successfully sent
                return false;
            }
        }

        /// <summary>
        /// Receives UDP messages.
        /// </summary>
        private void ReceiveUDP()
        {
            try
            {
                // Keep receiving UDP messages until the thread is aborted
                while(true)
                {
                    // Receive next UDP message
                    byte[] buffer = Receiver.Receive(ref ReceiverEndPoint);

                    // Process UDP message
                    if(buffer.Length > 0)
                    {
                        // Extract the message
                        string message = Encoding.ASCII.GetString(buffer);
                        
                        // Check the message
                        if(!CheckHexString(message))
                        {
                            Log("[INV MSG ] " + message);
                        }
                        // Add the message to the send queue if it is valid
                        else
                        {
                            lock(SendQueue)
                            {
                                SendQueue.Add(message);
                            }
                        }
                    }
                }
            }
            catch(Exception e)
            {
                // Log errors
                if(e.GetType() != typeof(System.Threading.ThreadAbortException))
                    Log("ERROR: " + e.Message);
            }
        }

        /// <summary>
        /// Sends a message via UDP to the configured host.
        /// </summary>
        /// <param name="message">The message to send.</param>
        private void SendUDP(string message)
        {
            try
            {
                // Send the UDP message
                byte[] buffer = Encoding.ASCII.GetBytes(message);
                Sender.Send(buffer, buffer.Length, SenderEndPoint);
            }
            catch(Exception e)
            {
                // Log errors
                Log("ERROR: " + e.Message);
            }
        }

        /// <summary>
        /// Checks if the given string is a valid hex string.
        /// </summary>
        /// <param name="hexString">The string to check.</param>
        /// <returns>True if the given string is a valid hex string, false otherwise.</returns>
        private bool CheckHexString(string hexString)
        {
            // Clean the string
            hexString = hexString.ToUpper().Replace(" ", "");

            // Check the length of the string
            if(hexString.Length % 2 != 0)
                return false;

            // Check each character of the string
            for(int i = 0; i < hexString.Length; i++)
                if(!((hexString[i] >= '0' && hexString[i] <= '9') || (hexString[i] >= 'A' && hexString[i] <= 'F')))
                    return false;

            // Return true if all checks have been passed
            return true;
        }

        /// <summary>
        /// Converts the given message to a hex string.
        /// </summary>
        /// <param name="message">The message to convert.</param>
        /// <returns>The given message as a hex string.</returns>
        private string ByteToHex(byte[] message)
        {
            // Prepare the string builder
            StringBuilder builder = new StringBuilder(message.Length * 3);

            // Convert each byte of the message
            foreach(byte data in message)
                builder.Append(Convert.ToString(data, 16).PadLeft(2, '0').PadRight(3, ' '));

            // Return the converted message
            return builder.ToString().ToUpper();
        }

        /// <summary>
        /// Converts the given message to a hex string.
        /// </summary>
        /// <param name="message">The message to convert.</param>
        /// <returns>The given message as a hex string.</returns>
        private string ByteToHex(byte message)
        {
            // Return the converted message
            return Convert.ToString(message, 16).PadLeft(2, '0').ToUpper();
        }

        /// <summary>
        /// Converts the given message to a byte array.
        /// </summary>
        /// <param name="message">The message to convert.</param>
        /// <returns>The given message as a byte array.</returns>
        private byte[] HexToByte(string message)
        {
            // Remove any spaces from the message
            message = message.Replace(" ", "");

            // Prepare the byte array
            byte[] buffer = new byte[message.Length / 2];

            // Convert each byte of the message
            for(int i = 0; i < message.Length; i += 2)
                buffer[i / 2] = (byte) Convert.ToByte(message.Substring(i, 2), 16);
            
            // Return the converted message
            return buffer;
        }
    }
}
