// Project: eBus Connector
//  Module: Service
//  Author: Thomas Ternyi (thomas@terenyi.at)

using System;
using System.Diagnostics;
using System.ServiceProcess;
using System.Threading;
using Microsoft.Win32;

namespace eBusConnector
{
    /// <summary>
    /// This class contains the service.
    /// </summary>
    public partial class Service : ServiceBase
    {
        /// <summary>
        /// The eBus Connector.
        /// </summary>
        private eBusConnector eBus;

        /// <summary>
        /// The service thread.
        /// </summary>
        private Thread ServiceThread;

        /// <summary>
        /// Indicates whether the service should run.
        /// </summary>
        private volatile bool RunService;

        /// <summary>
        /// Constructs a new service.
        /// </summary>
        public Service()
        {
            this.ServiceName = "eBusConnector";
        }

        /// <summary>
        /// Starts the service.
        /// </summary>
        /// <param name="args">Command line parameters.</param>
        protected override void OnStart(string[] args)
        {
            try
            {
                // Read the service settings
                string serialPort = Registry.GetValue("HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\eBusConnector", "SerialPort", "").ToString();
                string localUDPPort = Registry.GetValue("HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\eBusConnector", "LocalUDPPort", "").ToString();
                string remoteUDPPort = Registry.GetValue("HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\eBusConnector", "RemoteUDPPort", "").ToString();
                string host = Registry.GetValue("HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\eBusConnector", "RemoteHost", "").ToString();
                bool listenWhileSending = (Registry.GetValue("HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\eBusConnector", "ListenWhileSending", "").ToString() == "1");

                // Log that the service is being started
                EventLog.WriteEntry("eBusConnector", "Starting eBus Connector...\n\nSerial port: " + serialPort + "\nLocal UDP port: " + localUDPPort + "\nRemote UDP port: " + remoteUDPPort + "\nRemote host: " + host + "\nListen while sending: " + (listenWhileSending ? "On" : "Off"), EventLogEntryType.Information);

                // Check the service settings
                try
                {
                    if(serialPort == "")
                        throw new Exception();

                    if(localUDPPort == "")
                        throw new Exception();

                    if(remoteUDPPort == "")
                        throw new Exception();

                    if(host == "")
                        throw new Exception();

                    if(int.Parse(localUDPPort) <= 0 || int.Parse(localUDPPort) > 65535)
                        throw new Exception();

                    if(int.Parse(remoteUDPPort) <= 0 || int.Parse(remoteUDPPort) > 65535)
                        throw new Exception();
                }
                catch(Exception)
                {
                    throw new Exception("Service not correctly configured.");
                }

                // Create and configure the eBus Connector
                eBus = new eBusConnector(serialPort, int.Parse(localUDPPort), int.Parse(remoteUDPPort), host, listenWhileSending, false);
                eBus.Log += this.Log;

                // Start the eBus Connector
                eBus.Start();

                // Start the service thread
                RunService = true;
                ServiceThread = new Thread(this.Run);
                ServiceThread.Start();

                // Log that the service has been started
                EventLog.WriteEntry("eBusConnector", "eBus Connector started.", EventLogEntryType.Information);
            }
            catch(Exception e)
            {
                // Log errors to the event log
                EventLog.WriteEntry("eBusConnector", "ERROR: " + e.Message, EventLogEntryType.Error);
            }
        }

        /// <summary>
        /// Stops the service.
        /// </summary>
        protected override void OnStop()
        {
            try
            {
                // Log that the service is being stopped
                EventLog.WriteEntry("eBusConnector", "Stopping eBus Connector...", EventLogEntryType.Information);

                // Stop the eBus Connector
                eBus.Stop();

                // Stop the service thread
                RunService = false;
                ServiceThread.Join();

                // Log that the service has been stopped
                EventLog.WriteEntry("eBusConnector", "eBus Connector stopped.", EventLogEntryType.Information);
            }
            catch(Exception e)
            {
                // Log errors to the event log
                EventLog.WriteEntry("eBusConnector", "ERROR: " + e.Message, EventLogEntryType.Error);
            }
        }

        /// <summary>
        /// Runs the service.
        /// </summary>
        private void Run()
        {
            // Do nothing (the eBus Connector is running and doing the work)
            while(RunService)
                Thread.Sleep(1000);
        }

        /// <summary>
        /// Logs a message.
        /// </summary>
        /// <param name="message">The message to log.</param>
        private void Log(string message)
        {
            // Only log error messages to the event log
            if(message.Contains("ERROR"))
                EventLog.WriteEntry("eBusConnector", message, EventLogEntryType.Error);
        }
    }
}
